//********************************************************************************
//*         Name:  coprocm_spi.c                                                 *               
//*                                                                              *
//*       Author:  Daniel F. Ramirez                                             *
//*                                                                              *
//*          Date:  06/13/2003                                                   *
//*                                                                              *
//*   Description:                                                               *
//*                                                                              *
//*                                                                              *
//*     This is a Master Math Co-Processor application that uses a PIC18F8720 to * 
//* send expressions to the embedded Math Co-Processor using the SPI interface,*
//* in order to compute floating point arithmetic expressions in RPN format.  It *
//* can be used for Sumo Robots that require floating point for navigation and   *
//* tracking of an oponent with minimal space, power and weight requirements are *
//* prime candidates for using this co-processor, since it is very compact.      *
//* The master can be a PIC or STAMP or any other micro-controller that requires *
//* a math co-processor and has at least 4 spare I/O pins. In addition to the    *
//* basic arithmetic functions, the math co-processor can also evaluate          *
//* expressions that contain trigonometric and scientific functions.             *
//*                                                                              *
//*      A communication interface is provided that allows the host to use high  *
//* speed asynchronous serial link or SPI or I2C Master to request floating point*
//* equation calculations or to pass it individual sensor readings, average      *
//* sensor readings, calibrating a sensor or to perform a sensor processing      *
//* algorithm using command message packets.  The commands available are         *
//* described below.  This header file provides the common Master and Slave      *
//* message formats and descriptors.                                             *
//*/*                                                                            *
//* The high priority ISR will invoke the "Z;" reset command via the serial      *
//* interface. Both the serial interface and the USART interface are supported by  *
//* defining the approprate compiler flag.                                       *
//*                                                                              *
//*                                                                              *
//*     This master controller can be used to communicate with a PC host or any  *
//* SPI, I2C or USART slave, in order to  give the microcontroller the capability*
//* to serve as an SPI or I2C slave controller in order to provide it the        *
//* floating point coprocessor functions.                                        *  
//* It can also be used for robotic applications where multiple controllers and  *
//* SPI devices are used, with a minimum of data and control lines needed to     *
//* connect them.  Each I2C device has its own unique address and only listens   *
//* to commands associated with it.                                              *
//*                                                                              *
//*     Future upgrades to this application will include implementing various    *
//* filters including FIR, FFR, Kalman and Median filters.  Both FFT and IFFT    *
//* functions and a Histogram function are also in the works. Advanced algorithms*
//* such as Vector Force Fields (VFF), Fuzzy Logic and Neural Networks are new   *
//* areas that may be explored using this application.   In addition, this       *
//* application may be able to take advantage of the new DSPIc Signal Processors *
//* (DSP's) become available from Microchip.                                     *
//*                                                                              *
//* Usage:                                                                       *
//*	The following Math Co-Processor Controller command messages are available:   *
//*                                                                              * 
//* calculate    - Evalueates the RPN expressions that are stored in the         *
//* message expression buffer.  Multiple expressions can be sent to the          *
//* co-processor for evaluation, if they are separated by a ";" semi-colon.      *
//* Large expressions can span multiple messages to the co-processor.            *
//*                                                                              * 
//* get_answers  - Get the results from the latest arithmetic operations         *
//* carried out during the last cycle. Up to 7 separate single precision floating*
//* point and 3 double precision  values may be returned during a single message *
//* transfer.                                                                    *
//*                                                                              *
//* clear_memory - Erases the contents of the Math co-processor's storage        * 
//* registers. These same registers are used for storing and retrieving temporary*
//* values. It clears floating point registers 0..MAX_FLOAT_REGISTERS.           *
//*                                                                              *
//* store   - Stores a floating point value to a selected floating point         *
//* register. The register ID range is 0..MAX_FLOAT_REGISTERS.                   *                                                            
//*                                                                              *
//* recall 	- Retrieves a floating point value from the selected floating        *
//* point register. The register ID range is 0..MAX_FLOAT_REGISTERS.             *                                                         
//*                                                                              *                                                                     
//*	GetStatus - Get the Math Co-processor status including current state, and    *  
//* any error conditions encountered i.e. division by 0, overflow, underflow.    *
//*                                                                              *
//* ResetCommand - Reset the DC motor Controller (brake the motor).  This        *
//* command can be used to implement an emergency stop function.                 *  
//*                                                                              *
//*                                                                              *                       
//*                                                                              *
//********************************************************************************
    
// Include file headers here

#include "flags.h"                                      // Global flags used for MPSIM debugging 
  
#ifdef PIC18C252     
//#include <p18c252.h>					// Register definitions for PIC18C252
#endif

#ifdef PIC18C452     
#include <p18c452.h>					// Register definitions for PIC18C452
#endif

#ifdef PIC18F452     
#include <p18f452.h>					// Register definitions for PIC18F452
#endif
  
#ifdef PIC18F8720     
#include <p18f8720.h>					// Register definitions for PIC18F8720
#endif
 
#ifdef PIC18F6620     
#include <p18f6620.h>
#endif
 
#include "UARTIntC.h"					// Maestro USART library functions
#include "delays.h"

#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <usart.h>                                      // USART library functions
#include <spi.h>                                        // SPI library functions
#include <pwm.h>                                        // PWM library functions
#include <timers.h>                                     // Timer library functions
#include <portb.h>                                      // PORTB library functions                                 
#include <usart.h>                                      // USART library function
#include <delays.h>                                     // High resolution delay functions
 
//#include "portutil.h"                                 // Port utility functions
#include "timeutil.h"
#include "stdio.h"                                      // Serial I/O support functions
#include "network.h"                                   	// SPI Networking header
#include "spimast.h"                                   	// SPI Master services 
#include "coproc.h"										// Include Common Math Co-processor header
#include "coprocm.h"									// Include Math Co-processor Services header
 
  
//********************************************************************************
// Constant Definitions
//********************************************************************************
#define CR 13

 
//#define TEST                                          	// Flag used for debug I/O
//#define SPI_TEST                                    // Test serial I/O with this application
//#define GARDENBOT                                     // Control a GardenBot robot
#define GEARBOT                                         // Control a GearBot

//#define SERIAL_INTERFACE								// Use high speed serial link
//#define I2C_INTERFACE									// Use I2C interface
//#define SPI_INTERFACE                                 // Use SPI interface

//#define MASTER_TEST_0                                 // Select the SPI Master Controller test #0
//#define MASTER_TEST_1                                 // Select the SPI Master Controller test #1
//#define MASTER_TEST_2                                 // Select the SPI Master Controller test #2
//#define MASTER_TEST_3                                 // Select the SPI Master Controller test #3
//#define MASTER_TEST_4                      

#define NUMBER_CYCLES 	100								// Number of PWM Cycles to be sent to H-Bridge

#define PAUSE_TIME      100	 					  		// Pause between PWM commands (Acceleration) 


#define StackSize		 5								// HP-45 stack (5 levels)
#define RegisterSize	10								// HP-45 Registers (1-9)

// Declare global constant definitions here

#define TRUE 1
#define FALSE 0

#define INPUT 1											 // Port input direction 
#define OUTPUT 0										 // Port output direction 

// SPI Interface 

// Port B

//#define TRACE0		PORTBbits.RB0						// Trace bit 0 for logic analyzer
//#define TRACE0_DIR	TRISBbits.TRISB0

#define BUSY_FLAG		PORTBbits.RB0				   	    // SPI Slave Co-processor busy flag
#define BUSY_FLAG_DIR	TRISBbits.TRISB0			   	    // SPI Slave Co-Processor busy flag Direction

// Port C

#define SPI_SCL_DIR 			TRISCbits.TRISC3 					// SPI Clock Direction (Master)
#define SPI_SI_DIR 				TRISCbits.TRISC4 					// SPI Input Direction (Master)
#define SPI_SO_DIR 				TRISCbits.TRISC5 					// SPI Output Direction (Master)
   
// Port E

//#define BUSY_FLAG		PORTEbits.RE2				   	    // SPI Slave Co-processor busy flag
//#define BUSY_FLAG_DIR	TRISEbits.TRISE2			   	    // SPI Slave Co-Processor busy flag Direction

//#define SPI_CS		PORTEbits.RE2				   	    // SPI Slave Co-processor Select
//#define SPI_CS_DIR	TRISEbits.TRISE2			   	    // SPI Slave Co-Processor Select Direction

// Port F

#define SPI_CS		PORTFbits.RF7				   	    // SPI Slave Co-processor Slave Select
#define SPI_CS_DIR	TRISFbits.TRISF7			   	    // SPI Slave Co-Processor Slave Select Direction

//#define SS			PORTFbits.RF7						// SPI Slave Co-processor Slave Select 
//#define SS_DIR		TRISFbits.TRISF7					// SPI Slave Co-processor Select Select Direction 


// Variables

int Status = 0;											// Status return code

byte ControllerState; 		// State of selected slave controller, range (0..255) 
											// 		0 => NO_STATES		No States    
											//		1 => INITIALIZED	The selected slave contoller is Initialzed 
											//		2 => DONE 	 	 	The selected slave contoller is done processing current command
											// 		3 => BUSY 	 		The selected slave contoller is busy processing current command 
											// 		4 => ERROR 	 		The selected slave contoller encountered an error processing the current command 

byte ControllerErrorCode; 	// Error code returned by the selected slave controller, range (0..255) 
											// 		0 => NO_ERRORS 		No error codes  
											// 		1 => FAILURE 	 	Failed to complete command
											//  	2 => SUCCESS  	 	Completed command successfully 
 
byte ControllerDiagnosticCode; 	// Diagnostic message code returned by the selected slave controller, range (0..255) 
											// 		0 => NO_DIAGNOSTICS 	No Diagnostic messages
											//  	1 => DIAGNOSTIC 1  	Diagnostic message 2 
											// 		2 => DIAGNOSTIC 2 	Diagnostic message 1
 
// Declare SPI Master Buffer variables here

byte SPI_Rx_Buffer[SPI_BUFFER_SIZE];                  	// SPI Receive Buffer
byte SPI_Tx_Buffer[SPI_BUFFER_SIZE];                  	// SPI Transmit Buffer
byte SPI_Start_Buffer=0;                         		// Start pointer for ring buffer
byte SPI_End_Buffer = 0;								// SPI End Buffer Index

byte SPI_Rx_Flag = FALSE;                        		// Indicates that data was received into the ring buffer
word SPI_Timeout = 0;                            		// SPI I/O Timeout counter
byte SPI_Timeout_Flag;                           		// SPI Timeout flag
word SPI_Number_Characters = 0;						 	// SPI Number of characters in buffer
int SPI_Communications_Error;                    		// SPI I/O error flag
                                                 		// -1 = SPI Input Bufer overrun error
                                                 		// -2 = SPI Communications timeout error
                                                 		// -3 = SPI buffer empty error

int SPI_Interrupt_Count = 0;                   	// Number of MSSP interrupts
int SPI_Data_In = 0;								// Data byte read from the SPI Master
int SPI_Data_Out = 0;							// Data byte sent to the SPI Master
int SPI_Status;									// SPI Status returned from routines
int MessageIndex = 0;							// Message ID number
unsigned char var;


// Initialize the Math Co-processor Commands

word SelectedCommand = 0;

typedef union message_union_type
{
    CALCULATE_TYPE CalculateCommand;							// Calculate command message
	GET_ANSWERS_TYPE GetAnswersCommand;							// Get answers command message
	CLEAR_MEMORY_TYPE CLearMemoryCommand;						// Clear memory command message
	STORE_TYPE StoreCommand;									// Store command message
	RECALL_TYPE RecallCommand;                          		// Recall command message	
 	GET_STATUS_COMMAND_TYPE GetStatusCommand;					// Get status message
	RESET_COMMAND_TYPE ResetCommand;					 		// Reset command message

	COMMAND_MESSAGE_TYPE SelectedCommand;						// Selected command	
	MESSAGE_TYPE Message;										// Unformatted message
} MESSAGE_UNION_TYPE;

// Define the Math Co-processor Commands message
MESSAGE_UNION_TYPE CoProcessorMessage;
  
#ifdef SINGLE_PRECISION
float Registers[MAX_FLOAT_REGISTERS];
float Answers[MAX_ANSWERS];
float TheResults[MAX_ANSWERS];
#else
double Registers[MAX_FLOAT_REGISTERS];
double Answers[MAX_ANSWERS];
double TheResults[MAX_ANSWERS];
#endif


char af1[BufferSize];    // Holds ascii representation of an IEEE floating point number for testing
   
#ifdef TEST
unsigned char RxBuffer[34];
unsigned char TxBuffer[34];
#endif

unsigned char chData;
unsigned char TheCharacterRead;								// Character read
unsigned char TheCharacterSent;								// Character sent
byte SPIStatus;

void low_isr(void);

 
// FUNCTION Prototypes

void main(void);
void SyncWithSlave(void);
byte GetByte(void);
void SendByte(byte TheData); 
void InitializeSPIMaster(void);
unsigned char ReadSPIByte( void );
unsigned char WriteSPIByte( unsigned char data_out );



// serial interrupt taken as low priority interrupt
#pragma code uart_int_service = 0x08
void uart_int_service(void)
{
	_asm	goto low_isr	_endasm
	
}
#pragma code

#pragma	interruptlow low_isr save=section(".tmpdata")
void low_isr(void)
{	
	// call of library module function, MUST
	//UARTIntISR();
}

void main(void)
{
 	
	// local variables for all test cases and their initialisation
 	unsigned char j = 0;
	unsigned char i  = 0;

	for (i=0; i<MAX_ANSWERS; i++)
	{
	 	TheResults[i] = 0.0;  
	 	Answers[i] = 0.0;  
	}

 	// local variables for all test cases and their initialisation

   	SetupTimers();                                     	// Setup timers and software
 
  	// Configure USART for 20 MHz, 9600 Baud, 8-Bits, No parity, Asynchronous
   
	InitializeUsart(); 

#ifdef SPI_BUFFER_INTERFACE

  	 // Initialize the SPI Master hardware
   
   	InitializeSPIMaster();

	printf("SPI Master Application ...\n");
 	
 	printf("Delay 5 seconds ...\r\n");
 	pause(5000);
 	printf("End of 5 second delay. \r\n");
	
    TheCharacterSent = 0;

#ifdef TEST_0

   	// Test #1
   	// Send lower case characters to the SPI Co-Processor and read them back
   	// the SPI Co-Processor has converted them to upper case.
   	
    
    while(1)
   	{
 
		//SPI_CS = ENABLE; 						// Assert chip select
 
        SPIStatus = WriteSPIByte(TheCharacterSent);	// Send a character to the SPI Slave Co-Processor
   
        pause(2); 							// Wait some time to allow it to be processed.
         										// We may want handshaking here...
 
        // Read the data sent from the SPI Slave
   	   	TheCharacterRead = ReadSPIByte();
  
        //pause(2); 							// Wait some time to allow it to be processed.
         										// We may want handshaking here...
                                       

        //SPI_CS = DISABLE; 						// Negate chip select

      	printf("Master: character Sent = ");
        dec(TheCharacterSent);
        printf(" Character Received = ");
        dec(TheCharacterRead);
        printf(" SPI Status = ");
        dec(SPIStatus);
        printf("\r\n");
        //pause(1000); 				  		// Wait some time to allow the operator to see the results

        TheCharacterSent++;
      //}

   }

#endif 

  
   	// Test #1
   	// Send lower case characters to the SPI Co-Processor and read them back
   	// the SPI Co-Processor has converted them to upper case.
   	
 	TheCharacterSent = 0;
 	MessageIndex = 0;
 	
 	for (i=0; i<SPI_BUFFER_SIZE; i++)
	{
 		CoProcessorMessage.Message.Data[i] = 0;            			// Clear the message  
 	}

    // Test code
    
 	for (i=0; i<SPI_BUFFER_SIZE; i++)
	{
 		SPI_Rx_Buffer[i] = 0;                  		// Clear SPI Receive Buffer 
 		//SPI_Tx_Buffer[i] = i;                  	// Initialize SPI Transmit Buffer to a ramp 
 		SPI_Tx_Buffer[i] = 0;      	// Initialize SPI Transmit Buffer to a ramp 
	}
 
    
 
    
 
	//SPI_CS = ENABLE; 								// Assert chip select
    //SPI_CS = DISABLE; 							// Negate chip select
    
    SelectedCommand = 0;

 
	// Mainline Logic for SPI Buffer Interface goes here ...
		
	// This is the section where user selected command messages are issued to the Math Co-Processor 
	// In this case High Level Math Co-ProceController Commands are issued here ....
	// These functions may be used to test all the available the Math Functions and evaluate algorithms.
			
 	// Use the following table for evaluating an expression from an embedded application...


 /*

// General Math Coprocessor functions

#define GOLD    'a'					// Gold function key    
#define ENTER	'^'                 // Enter key  
#define ROLL    '|'					// Roll key   
#define SWAP    '~'					// Swap X with Y key  
#define STORE   '>'					// Store X into register N, N=1..9   
#define RECALL  '<'					// Recall register N, N=1..9 into X  
#define LASTX   '='                 // Recall register Last X into X

// Math Operators 

#define PLUS	'+'					// "+" Operator
#define MINUS	'-'					// "-" Operator
#define TIMES	'*'					// "*" Operator
#define DIVIDE	'/'					// "/" Operator

#define PERIOD	'.'					// "." Symbol
#define SEMICOLON	';'				// ";" Symbol

// Math functions

// Trigonometric Functions

#define SIN		'A'					// Sine function
#define ASIN  	'B'					// Tangent function
#define COS		'C'					// Cosine function
#define ACOS  	'D'					// Arc Cosine function
#define TAN		'E'					// Tangent function
#define ATAN	'F'					// Arc Tangent function
#define ATAN2	'G'					// Arc Tangent Two function
 
// Scientific Functions

#define SQRT	'H'					// Square root function
#define SQR		'I'					// Square function
#define EXP		'J'					// Exponential function E**X
#define LOG		'K' 				// Log function
#define LOG10	'L' 				// Log base 10 function
#define POWER	'M' 				// X**N  function
#define MAX		'N'                 // Maximum of x and y
#define MIN		'O'                 // Minimum of x and y
#define ABS		'P'                 // Absolute value of x
#define SIGN	'Q'                 // Sign of x and y

// Special functions

#define RAD 	'R'					// Convert to radians
#define DEG 	'S'					// Convert to degrees
#define FIX 	'T'					// Rounds x towards 0
#define FLOOR 	'U'					// Rounds x towards - Infinity
#define CEIL 	'V'					// Rounds x towards + Infinity
#define ROUND 	'W'					// Round x to nearest integer
#define PRECISION 	'X'				// Specify maximum number of digits precision up to machine precision used for display purposes
#define MOD		'Y'					// Mod Function identical to C mod function 
*/

   	// Test data for the math coprocessor application "^" is the "ENTER" key.  Use debugger to look at the results 
    // Second serial port for I/O to Windows Hyperterminal. 
   
 
	//********************************************************************************************
 	//* Send various arithmetic expressions to the Math Co-processor for evaluation
	//********************************************************************************************

    while (1)
    {
 
 		 // Test 1 - Send the coprocessor command message that evaluates the RPN expression: sqrt(25.0)*3.0 + 2  = 17.0
 
		SendCalculateCommand ("25.0H^3*^2+;");
        //pause(10);
 		//  Get Results into the floating point Results array where TheResults[0] is the answer to the first expression, etc. 
  		GetResults(TheResults);

 		 // Test 2 - Send the coprocessor command message that evaluates the RPN expression: sin(radians(30)) = 0.5
 		SendCalculateCommand ("30.0RA;");
        //pause(10);
 		//  Get Results into the floating point Results array where TheResults[0] is the answer to the first expression, etc. 
  		GetResults(TheResults);


 		 // Test 3 - Send the coprocessor command message that evaluates the RPN expression: (3+4)/(8+6) = 0.5
 		SendCalculateCommand ("3^4+8^6+/");
        //pause(10);
  		//  Get Results into the floating point Results array where TheResults[0] is the answer to the first expression, etc. 
  		GetResults(TheResults);

	 	// Test 4 - Send the coprocessor command message that evaluates the RPN expression: (8-2)*(6+3)= 54.0
 		SendCalculateCommand ("8.0^2.0-^6.0^3.0+*");
        //pause(10);
		//  Get Results into the floating point Results array where TheResults[0] is the answer to the first expression, etc. 
  		GetResults(TheResults);

	 	// Test 5 - Send the coprocessor command message that evaluates multiple RPN expressions from 
        //          Test 1 and Test 3: combined.
 		SendCalculateCommand ("25.0H^3*^2+;3^4+8^6+/;");
        //pause(10);
 		//  Get Results into the floating point Results array where TheResults[0] is the answer to the first expression, etc. 
  		GetResults(TheResults);

	
#ifdef NEW_CODE
  			
  		// Send Get Status Command();
 		SendGetStatusCommand();

  		// Send the Clear Memory Command message that erases the floating point registers
		SendClearMemoryCommand();
					
 		// Send the Store Command message that saves a floating point value to a selected float register
		SendStoreCommand();
					
  		// Send the Recall Command message that  Retrieves a floating point value from a 
  		// selected float register.	
  		SendRecallCommand();
   
   		// Send the Filter command message that performs selected filter 
  		// operation (FIR, IFR)

		SendFilterCommand();
 				
   		// Send the FFT command message that performs the selected 
  		//FFT operation on the specified data.

		SendFFTCommand();
 				
   		// Send the Inverse Filter command message that performs an inverse
  		// FFT operation on the specified data.

		SendIFFTCommand();
 				
   		// Send the Kalman Filter command message that performs a Kalman
  		// Filter on the specified data

		SendKalmanFilterCommand();
 				
   		// Send the Median Filter command message that performs a Median filter 
  		//operation on the specified data.

		SendMedianFilterCommand();
 				
   		// Send the Histogram command messagePerforms a Histogram operation

		SendHistogramCommand();
 				
   		// Process the Sort command message that performs the selected Sort operation 
  		//on the specified data

		SendSortCommand();
 
		
 		//**********************************************************************************
 		// Send and process the Get Status command response message from the Slave
		//**********************************************************************************
 
 		SendGetStatusCommand();
  	
 		//**********************************************************************************
 		// Send and process the Reset Command response message from the Slave
		//**********************************************************************************
 
 		//SendResetCommand();
#endif	
			
	}
		
#endif
	
}

//**********************************************************************************
//* SendSPIMessage - Sends a command message to the SPI Slave and stores the data received 
//* into the SPI_Rx_Buffer so that it may be processed by Reveive Message. 
//**********************************************************************************
byte SendSPIMessage(void)
{
    int i;														// Loop index 		
	byte SendSPIMessageStatus = TRUE;								// Status return

 
#ifdef TEST
 
	// Send a test ramp that spans 0..SPI_BUFFER_SIZE
	
	for (i=0; i<SPI_BUFFER_SIZE; i++)
	{
		CoProcessorMessage.Message.Data[i] = i;
  	}	
  	 	
#endif	
	
  	
 	// Copy the Ramp test message from the SPI Transmit Buffer to the 
 	// Coprocessor Message for test
    			
  	//memcpy( SPI_Tx_Buffer+TX_DATA_OFFSET, &CoProcessorMessage.Message.Data, SPI_BUFFER_SIZE);  

   	for (i=0; i<SPI_BUFFER_SIZE; i++)
	{
  			 
		SPI_Tx_Buffer[i] =  CoProcessorMessage.Message.Data[i];
	}
			
	// Wait until the slave is not busy

	while(BUSY_FLAG);
 
	
	//**********************************************************************************
	// Send an SPI message command here
	//**********************************************************************************

	for (i=0; i<SPI_BUFFER_SIZE; i++)
  	{
 
   		SPI_CS = ENABLE; 						// Assert chip select
        SPIStatus = WriteSPIByte(SPI_Tx_Buffer[i]);	// Send a character to the SPI Slave Co-Processor
        SPI_CS = DISABLE; 						// Negate chip select

        pause(2); 							// Wait some time to allow it to be processed.
         										// We may want handshaking here...
 
        // Read the dummy or status data sent back from the SPI Slave 
     	SPI_CS = ENABLE; 						// Assert chip select
  	   	TheCharacterRead = ReadSPIByte();
        SPI_CS = DISABLE; 						// Negate chip select
   
        pause(2); 							// Wait some time to allow it to be processed.
         										// We may want handshaking here...                                     

#ifdef TEST
      	printf("Master: Send: i = ");
        dec(i);
        printf("  SPI_Tx_Buffer[i] = ");
        dec(SPI_Tx_Buffer[i]);
        printf("  ");
        putch((char)SPI_Tx_Buffer[i]);
        printf(" SPI_Rx_Buffer = ");
        dec(SPI_Rx_Buffer[i]);
        printf("  ");
        putch((char)SPI_Rx_Buffer[i]);
        printf("\r\n");
        //pause(1000); 				  		// Wait some time to allow the operator to see the results
#endif 
   
 	
 	}
	
#ifdef TEST		
	

	// Display contents of SPI Buffers 	
			 
	for (i=0; i<SPI_BUFFER_SIZE; i++)
	{
		printf("Master: i = ");
		dec(i);
		printf("  SPI_Tx_Buffer = ");
		dec(SPI_Tx_Buffer[i]);
			
		printf("  SPI_Rx_Buffer = ");
		dec(SPI_Rx_Buffer[i]);
        printf("  ");
        send_byte(SPI_Rx_Buffer[i]);
			
		printf("\r\n");
			
		//pause(1000);
	}
	
	printf("\r\n\r\n");
 

#endif
		
	return SendSPIMessageStatus;
}

//**********************************************************************************
//* ReceiveSPIMessage - Receives and processes the message that is in the SPI_Rx_Buffer 
//* which was simultaneously sent back from the SPI Slave Controller.  The message may
//* contain data or status information.
//**********************************************************************************
byte ReceiveSPIMessage(void)
{
	int i; 															// Loop index 
 	byte ReceiveSPIMessageStatus = TRUE;							// Status return

		 				
  	// Copy the entire Co-Processor message received from the SPI Slave
   	// to local storage for processing.  It may include sensor data, telemetry and status 
   	// information and is received concurrently with each message that the Master SPI controller sends.
   			
  	//memcpy(&CoProcessorMessage.Message.Data, (char *) (SPI_Rx_Buffer+RX_DATA_OFFSET), SPI_BUFFER_SIZE);  
  	
   			
   	for (i=0; i<SPI_BUFFER_SIZE; i++)
	{
  			
		CoProcessorMessage.Message.Data[i] = SPI_Rx_Buffer[i];
	}
     		



	// Clear the last message
   	for (i=0; i<SPI_BUFFER_SIZE; i++)
	{
  			
		CoProcessorMessage.Message.Data[i] = 0;
	}

	// Flush the Rx Buffer
	for (i=0; i<SPI_BUFFER_SIZE; i++)
	{
 		SPI_Rx_Buffer[i] = 0;                  		// Clear SPI Receive Buffer 
	}
	
	//**********************************************************************************
	// Receive an SPI message from the Math Co-processor and process it  here
	//**********************************************************************************
 
	// Wait until the slave is not busy

	//while(BUSY_FLAG);
 
 
	for (i=0; i<SPI_BUFFER_SIZE; i++)
  	{
  
  		SPI_CS = ENABLE; 						// Assert chip select
      	SPIStatus = WriteSPIByte(TheCharacterSent);	// Send a dummy character to the Slave Co-Processor
        SPI_CS = DISABLE; 						// Negate chip select

        pause(2); 							// Wait some time to allow it to be processed.
         										// We may want handshaking here...
 
        // Read the data sent back from the SPI Slave 
  		SPI_CS = ENABLE; 						// Assert chip select
  	   	SPI_Rx_Buffer[i] = ReadSPIByte();
        SPI_CS = DISABLE; 						// Negate chip select
 
        pause(2); 							// Wait some time to allow it to be processed.
         										// We may want handshaking here...
                                       

#ifdef TEST
      	printf("Master: Receive: i = ");
        dec(i);
        printf("  SPI_Rx_Buffer[i] = ");
        dec(SPI_Rx_Buffer[i]);
        printf("  ");
        putch((char)SPI_Rx_Buffer[i]);
        printf("\r\n");
        //pause(1000); 				  		// Wait some time to allow the operator to see the results
#endif 
   
 	
 	}

 		
#ifdef TEST	
	

	// Display contents of SPI Buffers 	
			 
	for (i=0; i<SPI_BUFFER_SIZE; i++)
	{
		printf("Master: i = ");
		dec(i);
		printf("  SPI_Tx_Buffer = ");
		dec(SPI_Tx_Buffer[i]);
			
		printf("  SPI_Rx_Buffer = ");
		dec(SPI_Rx_Buffer[i]);
			
		printf("\r\n");
			
		//pause(1000);
	}
	
	printf("\r\n\r\n");
/*	

	// Display contents of SPI message
	
	for (i=0; i<SPI_BUFFER_SIZE; i++)
	{
		printf("Master: i = ");
		dec(i);
		printf("  CoProcessorMessage.Message.Data = ");
		dec(CoProcessorMessage.Message.Data[i]);
		printf("\r\n");
	}
*/

#endif

 	 				
  	// Copy the entire message received from the math coprocessor 
   	// to local storage for processing.  It may include sensor data, telemetry and status 
   	// information and is received concurrently with each message that the Master SPI controller sends.
   			
  	//memcpy(&CoProcessorMessage.Message.Data, SPI_Rx_Buffer+RX_DATA_OFFSET, SPI_BUFFER_SIZE);  
  			  
  			
   	for (i=0; i<SPI_BUFFER_SIZE; i++)
	{
  			
		CoProcessorMessage.Message.Data[i] = SPI_Rx_Buffer[i+1];
	}
    		
 	//**********************************************************************************************
 	// Display the message received from the SPI Slave here
 
 
    // Process the message received  from the Math Co-processor Slave here ...
			        	
 	switch (CoProcessorMessage.SelectedCommand.Command)
 	{
		case CALCULATE_COMMAND:
		{
 			// Process the Calculate command message that evaluates RPN expression
 			ReceiveCalculateCommand();
 				
			break;
		} 
			
		case GET_ANSWERS_COMMAND:
		{
		
 			// Process Get Answers command message in order to return the floating point answers to the master
  			ReceiveGetAnswers();
  					
  			break;
		}
			
		case CLEAR_MEMORY_COMMAND:
		{
  			// Process the Clear Memory Command message that erases the floating point registers
			ReceiveClearMemoryCommand();
					
 			break;
		}
			
		case STORE_COMMAND:
		{

 			// Process the Store Command message that saves a floating point value to a selected float register
			ReceiveStoreCommand();
					
   			break;
		}
			
		case RECALL_COMMAND:
		{

  			// Process the Recall Command message that  Retrieves a floating point value from a 
  			// selected float register.	
  			ReceiveRecallCommand();
   			break;
		}
#ifdef NEW_CODE
		case FILTER_COMMAND:
		{
  			// Process the Filter command message that performs selected filter 
  			// operation (FIR, IFR)

			ReceiveFilterCommand();
    		break;
		}
				
		case FFT_COMMAND:
		{
  			// Process the FFT command message that performs the selected 
  			//FFT operation on the specified data.

			ReceiveFFTCommand();
    		break;
		}
				
		case IFFT_COMMAND:
		{
  			// Process the Inverse Filter command message that performs an inverse
  			// FFT operation on the specified data.

			ReceiveIFFTCommand();
    		break;
		}
				
		case KALMAN_FILTER_COMMAND:
		{
  			// Process the Kalman Filter command message that performs a Kalman
  			// Filter on the specified data

			ReceiveKalmanFilterCommand();
    		break;
		}
				
		case MEDIAN_FILTER_COMMAND:
		{
  			// Process the Median Filter command message that performs a Median filter 
  			//operation on the specified data.

			ReceiveMedianFilterCommand();
    		break;
		}
				
		case HISTOGRAM_COMMAND:
		{	
  			// Process the Histogram command messagePerforms a Histogram operation

			ReceiveHistogramCommand();
    		break;
		}
				
		case SORT_COMMAND:
		{
  			// Process the Sort command message that performs the selected Sort operation 
  			//on the specified data

			ReceiveSortCommand();
    		break;
		}
#endif			
		case GET_STATUS_COMMAND:
		{
			//**********************************************************************************
 			// Receive and process the Get Status command response message from the Slave
			//**********************************************************************************
 
 			ReceiveGetStatusResponse();
   			break;
		}
 	
		case RESET_COMMAND:
		{
			//**********************************************************************************
 			// Receive and process the Reset Command response message from the Slave
			//**********************************************************************************
 
 			ReceiveResetResponse();
   			break;
		}	

	}
		
	
 	//**********************************************************************************************
	
	
	//pause(2000);
	 	
 	// Generate next command 
		
	//SelectedCommand++;   
	//SelectedCommand = SelectedCommand % 6;
 
    
#ifdef SPI_NETWORK_INTERFACE
	// Mainline Logic for SPI Network Interface goes here ...
		
	SPI_CS = ENABLE; 						// Assert chip select
 
    putsSPI(SPI_Tx_Buffer);					// Send the message stored in the SPI_Tx_Buffer to the 
                                  				// SPI Slave Co-Processor
 
    //pause(5); 								// Wait some time to allow it to be processed.
         										// We may want handshaking here...
 
    // Read the message data sent from the SPI Slave
   	getsSPI(SPI_Rx_Buffer, sizeof(SPI_Rx_Buffer));
 
   	   	
    SPI_CS = DISABLE; 						// Negate chip select
  
#endif

   return ReceiveSPIMessageStatus;
}
 	
//**********************************************************************************
//* CheckControllerState - Check the status of the the current controller to see if
//* the command has been executed correctly.  Check the following Controller state
//* variables: 
//*		ControllerState	{ NO_STATES=0, INITIALIZED=1, DONE=2, BUSY=3, ERROR=4 }
//* 	ControllerErrorCode  { NO_ERRORS=0, FAILURE=1, SUCCESS=2 }	 
//* 	ControllerDiagnosticCode { NO_DIAGNOSTICS=0, DIAGNOSTIC1=1, DIAGNOSTIC2=2 }    
//**********************************************************************************
 byte CheckControllerState(void)
 {

	byte CheckControllerStateStatus = NO_STATES;
	
	// Start out with Controller State set to "NO_STATES" and continue to send and
	// receive Get Status messages until either the controller state changes to "DONE" or
	// gets an error state.
	
    ControllerState =  NO_STATES;
    
	SendGetStatusCommand();
	ReceiveGetStatusResponse(); 
	
	while (ControllerState != DONE)
	{
		  SendGetStatusCommand();
		  ReceiveGetStatusResponse();
#ifdef TEST	  
		  printf("ControllerState = ");
		  dec(ControllerState);
		  printf("  ControllerErrorCode = ");
		  dec(ControllerErrorCode);
		  printf("  ControllerDiagnosticCode = ");
		  dec(ControllerDiagnosticCode);
		  printf("\r\n");
		  
#endif	
	  
		  CheckControllerStateStatus = ControllerState;
		  if (ControllerState == ERROR)
		  {	  
		  		//printf("Error # ");
		  		//dec(ControllerErrorCode);
		  		//printf(" was encountered by the Math Co-processor!!\r\n");
		  		
		  		break;
		  }
	}
			

#ifdef TEST	  
	if (ControllerState == DONE)
	{
		printf("Math Co-processor Command completed successfully!\r\n");
	}
#endif 
	return CheckControllerStateStatus;
	
}

//**********************************************************************************
//**********************************************************************************
//**********************************************************************************
//**********************************************************************************

// Send command messages to Math Co-processor here...
	 		
//**********************************************************************************
//* SendCalculateCommand - Send the Calculate command to the Math Co-processor so  * 
//* that it can Evalueate the RPN expressions that are stored in the message       *
//* expression buffer. Return a status byte indicating the outcome of the Calculate*
//* Command process.                                                               *
//**********************************************************************************
byte SendCalculateCommand(const rom char *Data)  
{

 	byte CalculateCommandStatus = TRUE;							// Status return
  	word CalculateCommandDelay = 2;								// Delay to allow the duty command to complete
	int i;														// Local loop index
 
	// Clear the SPI send and receive buffers
	//FlushSPIBuffers();

	// Build the Calculate Command 
	
 	CoProcessorMessage.CalculateCommand.MessageID = MessageIndex++;		// Message ID range (0000 - FFFF) Hex
	CoProcessorMessage.CalculateCommand.Command = CALCULATE_COMMAND;	// Calculate command evaluates RPN expression
 
 	// Initialize the message buffer with an RPN expression to be evaluated by the Math co-processor 
 
 	//strcpy(&CoProcessorMessage.CalculateCommand.Buffer, (rom char *) Data);	 

	//for (i=0; i<strlen((rom char *) Data); i++)
	for (i=0; i<MESSAGE_BUFFER_SIZE; i++)
	{
		CoProcessorMessage.CalculateCommand.Buffer[i] = Data[i];
	}

	// Send the Calculate command message that evaluates RPN expression...

 		
	
/*	
	// *****************Missing logic to enter the expression from keyboard or directly from algorithm *****************
	for (i=0; i<MESSAGE_BUFFER_SIZE;i++)
	{
		CoProcessorMessage.CalculateCommand.Buffer[i] = 0;
	}
*/	
	
#ifndef TEST		

    printf("Master: Send Calculate Command  Command...\r\n");
  	printf("MessageID = ");
   	dec(CoProcessorMessage.CalculateCommand.MessageID);
/*
   	printf(",  Buffer = !");
   	for (i=0; i<20;i++)
   	{
   		printf("i = ");
   		dec(i);
   		printf("Buffer = ");
   		dec(CoProcessorMessage.CalculateCommand.Buffer[i]);
   		printf("\r\n");
   	}
   	//string(CoProcessorMessage.CalculateCommand.Buffer);
*/
   	printf("! \r\n\r\n");
#endif 

	// Send the command message to the SPI Slave Controller
	SendSPIMessage();

	// Give enough time for the command to complete 
	// ???????????????
	
	pause(CalculateCommandDelay);
	
	// Receive and process the response message from the SPI Slave Controller
	ReceiveSPIMessage();
		
 	// Check the status of the the current controller to see if the command has been 
	// executed correctly.
			
	//while (CheckControllerState() != DONE);
 
	return CalculateCommandStatus;
} 

//**********************************************************************************
//* SendGetAnswers  - Send command to the Math Co-procesor to get the results from *
//* the latest arithmetic operations carried out during the last cycle. Up to 7    *
//* separate single precision floating point and 3 double precision  values may be *
//* returned during a single message transfer.  Return a status byte indicating the*
//* outcome of the Send Get Answers command.                                       *
//**********************************************************************************
 
byte SendGetAnswers(void)
{
 	byte GetAnswersStatus = TRUE;								// Status return
 	word GetAnswersDelay = 20;									// Delay to allow the duty command to complete
 	byte i;														// Local loop index
 	
		
	// Build the Get Answers Command message 
			
	CoProcessorMessage.GetAnswersCommand.MessageID = MessageIndex++;				// Message ID range (0000 - FFFF) Hex 
  	CoProcessorMessage.GetAnswersCommand.Command = GET_ANSWERS_COMMAND;  			// Gets the floating point answers (see enumeration
               			                 									        // declaration above.)                                 	 
	CoProcessorMessage.GetAnswersCommand.NumberOfAnswers = 0;		

	// Clear the SPI send and receive buffers
	//FlushSPIBuffers();
 
#ifndef TEST
    printf("Master: Sending Get Answers Command: ");
  	printf("  Message ID = ");
    dec(CoProcessorMessage.GetAnswersCommand.MessageID);
    printf("\r\n\r\n");
#endif


	// Send the command message to the SPI Slave Controller to execute the command
	SendSPIMessage();
  
	// Give enough time for the command to complete
	// ???????????????
	
	pause(GetAnswersDelay);
	
	// Receive and process the response message from the SPI Slave Controller
	ReceiveSPIMessage();
	
	
 	// Check the status of the the current controller to see if the command has been 
	// executed correctly.
			
	//while (CheckControllerState() != DONE);
 	
	return GetAnswersStatus;
}
			
//**********************************************************************************
//* SendClearMemoryCommand - Send the clear memory command that erases the contents*
//* of the Math co-processor's storage registers. These same registers are used for*
//* storing and retrieving temporary values. It clears floating point registers    *
//* 0..MAX_FLOAT_REGISTERS. Return a status byte indicating the outcome of the     *
//* send clear memory command.                                                     *
//**********************************************************************************
byte SendClearMemoryCommand(void)
{
 	byte ClearMemoryCommandStatus = TRUE;										// Status return
 	word ClearMemoryCommandDelay = 2;											// Delay to allow the duty command to complete

 	// Clear the SPI send and receive buffers
	//FlushSPIBuffers();
 				
  	// Build the Clear Memory Command message
 	
	CoProcessorMessage.CLearMemoryCommand.MessageID = MessageIndex++; 			// Message ID (0000 - FFFF) Hex
	CoProcessorMessage.CLearMemoryCommand.Command = CLEAR_MEMORY_COMMAND;		// Clears the floating point scratch registers (see enumeration
        		                        										// declaration above.)  
 
#ifndef TEST		

    printf("Master: Send Clear Memory Command...\r\n");
  	printf("MessageID = ");
    dec(CoProcessorMessage.CLearMemoryCommand.MessageID);
    printf("  Command = CLEAR_MEMORY_COMMAND");
    printf("\r\n\r\n");
#endif 
	
	// Send the command message to the SPI Slave Controller
	SendSPIMessage();

	// Give enough time for the command to complete
	
	pause(ClearMemoryCommandDelay);
	//??????????????????????????

	// Receive and process the response message from the SPI Slave Controller
	ReceiveSPIMessage();
	
	
 	// Check the status of the the current controller to see if the command has been 
	// executed correctly.
			
	//while (CheckControllerState() != DONE);
 
	return ClearMemoryCommandStatus;
}

//**********************************************************************************
//* SendStoreCommand - Sends store command that stores a floating point value to a *
//* selected floating point register.     The register ID range is                 *             
//* 0..MAX_FLOAT_REGISTERS.                                                        *
//**********************************************************************************
byte SendStoreCommand(void)
{
 	byte StoreCommandStatus = TRUE;							// Status return
 	word StoreCommandDelay = 2;								// Delay to allow the Store command to complete

 	// Clear the SPI send and receive buffers
	//FlushSPIBuffers();
 
 	// Build Store Command message
 		
	CoProcessorMessage.StoreCommand.MessageID = MessageIndex++; 	// Message ID (0000 - FFFF) Hex

  	CoProcessorMessage.StoreCommand.Command = STORE_COMMAND;		// Stores a floating point value to a selected float register
  
  	     			                           						// (see enumeration declaration above.)  
 		
	// *****************Missing logic to enter the register ID and value
	// from keyboard or directly from algorithm *****************
	// ???????????????

 	// Test #2 - Send up to MAX_FLOAT_REGISTER values to be stored in the Math 
 	// Co-processor's memory. First, initialize the message register ID and 
 	// register value with test data for input to the Math Co-processor
 
 	CoProcessorMessage.StoreCommand.RegisterID = 0; 			// Selected register range (0..MAX_FLOAT_REGISTERS)  
 	CoProcessorMessage.StoreCommand.RegisterValue = 3.14159263; // Selected register range (0..MAX_FLOAT_REGISTERS)  
 	//CoProcessorMessage.StoreCommand.RegisterID = 1; 			// Selected register range (0..MAX_FLOAT_REGISTERS)  
 	//CoProcessorMessage.StoreCommand.RegisterValue = 2.781828; // Selected register range (0..MAX_FLOAT_REGISTERS)  

#ifndef TEST		

    printf("Master: Store Command...\r\n");
  	printf("MessageID = ");
    dec(CoProcessorMessage.StoreCommand.MessageID );
    printf("  Register ID = ");
    dec(CoProcessorMessage.StoreCommand.RegisterID);   
    printf(",  Register Value = ");
    dec(CoProcessorMessage.StoreCommand.RegisterValue);
	DoubleToAscii(CoProcessorMessage.StoreCommand.RegisterValue, af1);
   	puts(af1);
    printf("\r\n\r\n");
#endif	
		
	// Send the command message to the SPI Slave Controller
	SendSPIMessage();

	// Give enough time for the command to complete
	
	pause(StoreCommandDelay);
	//???????????????????????????????

	// Receive and process the response message from the SPI Slave Controller
	ReceiveSPIMessage();
	
		
 	// Check the status of the the current controller to see if the command has been 
	// executed correctly.
			
	//while (CheckControllerState() != DONE);
 
	return StoreCommandStatus;
}
  					
//**********************************************************************************
//* SendRecallCommand - Sends the recall command that retrieves a floating point   *
//* value from the selected floating point register. The register ID range is      *
//* 0..MAX_FLOAT_REGISTERS.                                                        *
//**********************************************************************************
byte SendRecallCommand(void)
{
 	byte RecallCommandStatus = TRUE;									// Status return
 	word RecallCommandDelay = 2;										// Delay to allow the duty command to complete

 	// Clear the SPI send and receive buffers
	//FlushSPIBuffers();
 				
  	// Build Recall Command message
 
 	CoProcessorMessage.RecallCommand.MessageID = MessageIndex++;   		// Message ID (0000 - FFFF) Hex

  	CoProcessorMessage.RecallCommand.Command = 	RECALL_COMMAND;  		// Retrieves a floating point value from a selected float register	
   	                                							    	// Command (see enumeration declaration above.)  
	// *****************Missing logic to enter the register ID and value
	// from keyboard or directly from algorithm *****************
   	//CoProcessorMessage.RecallCommand.RegisterID = 0;
	// ???????????????

 	// Test #3 - Recall up to MAX_FLOAT_REGISTERS values from the Math 
	// Co-processors memory.
	
 	// Initialize the message register ID with test data 
 	// for input to the Math Co-processor
 
 	CoProcessorMessage.RecallCommand.RegisterID = 0;
 	//CoProcessorMessage.RecallCommand.RegisterID = 1;
	//CoProcessorMessage.RecallCommand.RegisterID = 2;
	//CoProcessorMessage.RecallCommand.RegisterID = 3;
	//CoProcessorMessage.RecallCommand.RegisterID = 4;
	//CoProcessorMessage.RecallCommand.RegisterID = 5;
	//CoProcessorMessage.RecallCommand.RegisterID = 6;
  					
#ifndef TEST		

    printf("Master: Send Recall Command...\r\n");
  	printf("MessageID = ");
  	dec(CoProcessorMessage.RecallCommand.MessageID);
    printf("  Register ID = ");
    dec(CoProcessorMessage.RecallCommand.RegisterID);   
    printf(",  Register Value = ");
    dec(CoProcessorMessage.RecallCommand.RegisterValue);
	DoubleToAscii(CoProcessorMessage.RecallCommand.RegisterValue, af1);
   	puts(af1);
    printf("\r\n\r\n");
#endif	
													 
	// Send the command message to the SPI Slave Controller
	SendSPIMessage();

	// Give enough time for the command to complete
	
	pause(RecallCommandDelay);
	//?????????????????????????

	// Receive and process the response message from the SPI Slave Controller
	ReceiveSPIMessage();
	
		
 	// Check the status of the the current controller to see if the command has been 
	// executed correctly.
			
	//while (CheckControllerState() != DONE);

	return RecallCommandStatus;
}


//**********************************************************************************
//* SendResetCommand - Send the Reset the SPI Slave Controller.  This command 
//* resets the selected controller.  It may be used to implement an emergency stop 
//* function using the PIC's watchdog timer. 
//**********************************************************************************
byte SendResetCommand(void)
{ 

	byte ResetCommandStatus;
 	word ResetCommandDelay = 2;											// Delay to allow the duty command to complete
	
	// Clear the SPI send and receive buffers
	//FlushSPIBuffers();

	// Build Reset Command Message
	
	CoProcessorMessage.ResetCommand.MessageID = MessageIndex++;	// Message ID range (0000 - FFFF) Hex
	CoProcessorMessage.ResetCommand.Command = RESET_COMMAND;		// Send Reset Command
   
#ifndef TEST
    printf("Master: Send Reset Command...");
  	printf("MessageID = ");
   	dec(CoProcessorMessage.ResetCommand.MessageID); 
   	printf("\r\n\r\n");
#endif 


	// Send the command message to the SPI Slave Controller
	SendSPIMessage();

	// Give enough time for the command to complete
	
	pause(ResetCommandDelay);
	//?????????????????????????????

	// Receive and process the response message from the SPI Slave Controller
	ReceiveSPIMessage();

	
 	// Check the status of the the current controller to see if the command has been 
	// executed correctly.
			
	//while (CheckControllerState() != DONE);
 
	return ResetCommandStatus;
	
}

//**********************************************************************************
//* SendGetStatusCommand - Request to get the latest status from the selected Slave*
//* controller.  The message includes the controller state, any errors detected and*
//* any diagnostic messages reported.                                              *
//**********************************************************************************
byte SendGetStatusCommand(void)
{ 

	byte GetStatusCommandStatus;
 	word GetStatusCommandDelay = 5;											// Delay to allow the duty command to complete
	
	// Clear the SPI send and receive buffers
	//FlushSPIBuffers();

	// Build Reset Command Message
	
	CoProcessorMessage.GetStatusCommand.MessageID = MessageIndex++;	// Message ID range (0000 - FFFF) Hex
	CoProcessorMessage.GetStatusCommand.Command = GET_STATUS_COMMAND;		// Send Get Status Command
   
#ifndef TEST
    printf("Master: Send Get Status Command...");
  	printf("MessageID = ");
   	dec(CoProcessorMessage.GetStatusCommand.MessageID);
  
   	printf("\r\n\r\n");
#endif 

	// Send the command message to the SPI Slave Controller
	SendSPIMessage();
	
	// Give enough time for the command to complete
	
	pause(GetStatusCommandDelay);
	//?????????????????????????????

	// Receive and process the response message from the SPI Slave Controller
	ReceiveSPIMessage();
	
		
 	// Check the status of the the current controller to see if the command has been 
	// executed correctly.
			
	//while (CheckControllerState() != DONE);
 	

	return GetStatusCommandStatus;
	
}
		
//**********************************************************************************
//**********************************************************************************
//**********************************************************************************
//**********************************************************************************
// Functions used to process responses from the Slave Math Co-processor  

// Receive Responses from Slave here...

//**********************************************************************************
//* ReceiveCalculateCommand - Evalueates the RPN expressions that are stored in    *
//* the message expression buffer.  Multiple expressions can be sent to the        *
//* co-processor for evaluation, if they are separated by a ";" semi-colon.  Large *
//* expressions can span multiple messages to the co-processor.                    *
//**********************************************************************************
byte ReceiveCalculateCommand(void)  
{
 	byte ReceiveCalculateCommandStatus = TRUE;							// Status return
 	byte i;
 	
 	// Parse and Evalueate the RPN expression contained in the co-processor message buffer ...
 	
 	
 	
#ifndef TEST
    printf("slave: Calculate Command:");
  	printf("  MessageID = ");
   	dec(CoProcessorMessage.CalculateCommand.MessageID);
   	printf("\r\n");
    printf("Buffer = !");
   	for (i=0; i<20;i++)
   	{
   		printf("i = ");
   		dec(i);
   		printf("Buffer = ");
   		dec(CoProcessorMessage.CalculateCommand.Buffer[i]);
   		printf("\r\n");
   	}

    printf("!");
   	printf("\r\n\r\n");
#endif

 				
	return ReceiveCalculateCommandStatus;
}

//**********************************************************************************
//* ReceiveGetAnswers - Get the results from the latest arithmetic operations      *
//* carried out during the last cycle. Up to 7 separate single precision floating  *
//* point or 3 double precision  values may be returned during a single message    *
//* transfer.                                                                      *
//**********************************************************************************
byte ReceiveGetAnswers(void)
{
 	byte ReceiveGetAnswersStatus = TRUE;								// Status return
   	byte AnswerID;									 		    // Local Answer Loop Index 
    				
  	// Send  latest answers from requested RPN expressions back to the master.

#ifndef TEST
    printf("master: Receive Get Answers Command: \r\n");
  	printf("  MessageID = ");
   	dec(CoProcessorMessage.GetAnswersCommand.MessageID);
    printf(", Number Of Answers = ");
    dec(CoProcessorMessage.GetAnswersCommand.NumberOfAnswers);  
    printf("\r\n");
#endif

	// Display all the answers in floating point format
	
 	//for (AnswerID = 0; AnswerID < CoProcessorMessage.GetAnswersCommand.NumberOfAnswers; AnswerID++)
 	for (AnswerID = 0; AnswerID < MAX_ANSWERS; AnswerID++)
	{	
		// Copy local answers to Answer message
		Answers[AnswerID] = CoProcessorMessage.GetAnswersCommand.Answers[AnswerID];
#ifdef TEST
			
  		printf("  Answer ID = ");
    	dec(AnswerID);
    	printf(", Answer = ");
		//DoubleToAscii(CoProcessorMessage.GetAnswersCommand.Answers[AnswerID], af1);
   		//puts(af1);
   		hex4(CoProcessorMessage.GetAnswersCommand.Answers[AnswerID]);	              	
    	printf("\r\n");
#endif

	}
		 
	return ReceiveGetAnswersStatus;
}
			
//**********************************************************************************
//* ReceiveClearMemoryCommand - Erases the contents of the Math co-processor's 
//* floating point storage registers. These same registers are used for storing and
//* retrieving temporary values and answers. It does this by setting floating point 
//* registers 0..MAX_FLOAT_REGISTERS to zero.  Registers 0..10 are reserved for answers.
//**********************************************************************************
byte ReceiveClearMemoryCommand(void)
{
 	byte ReceiveClearMemoryCommandStatus = TRUE;							// Status return
    byte RegisterID;									 		    // Local Loop Index
	
#ifndef TEST

 	// Display the current register values   
 
  	printf("  MessageID = ");
   	dec(CoProcessorMessage.CLearMemoryCommand.MessageID);
    printf("master: Clear Memory Command Completed.");
   	printf("  MessageID = ");
   	dec(CoProcessorMessage.CLearMemoryCommand.MessageID);
   printf("\r\n\r\n");
#endif

	return ReceiveClearMemoryCommandStatus;
}

//**********************************************************************************
//* ReceiveStoreCommand - Stores a floating point value to a selected floating     *
//* point register.     The register ID range is 0..MAX_FLOAT_REGISTERS.           *
//**********************************************************************************
byte ReceiveStoreCommand(void)
{
 	byte ReceiveStoreCommandStatus = TRUE;										// Status return
			
 	// Store the value into the selected register.
  					
 	Registers[CoProcessorMessage.StoreCommand.RegisterID] = CoProcessorMessage.StoreCommand.RegisterValue;
  	 
#ifndef TEST
    printf("master: Store Command:");
  	printf("  MessageID = ");
   	dec(CoProcessorMessage.GetStatusCommand.MessageID);
   	printf("  Register ID = ");
   	dec(CoProcessorMessage.StoreCommand.RegisterID);
    printf("Register Value = ");
    hex4(Registers[CoProcessorMessage.StoreCommand.RegisterID]);
    printf("\r\n\r\n");
#endif 

	return ReceiveStoreCommandStatus;
}
  					
//**********************************************************************************
//* ReceiveRecallCommand - Retrieves a floating point value from the selected      *
//* floating point register. The register ID range is 0..MAX_FLOAT_REGISTERS.      *                                                               *
//**********************************************************************************
byte ReceiveRecallCommand(void)
{
 	byte ReceiveRecallCommandStatus = TRUE;									// Status return
 	
    // Accept number of samples xxxx to be used for calibration and data
   	// collection used for reading the DC Motor Control Pots, where the 
   	// number of samples range 0000..FFFF.
	
  	CoProcessorMessage.RecallCommand.RegisterValue = Registers[CoProcessorMessage.RecallCommand.RegisterID];
  	
#ifndef TEST
    printf("master: Recall Command:");
  	printf("  MessageID = ");
   	dec(CoProcessorMessage.RecallCommand.MessageID);
   	printf("  Register ID = ");
   	dec(CoProcessorMessage.RecallCommand.RegisterID);
    printf("Register Value = ");
    hex4(CoProcessorMessage.RecallCommand.RegisterValue);
    printf("\r\n\r\n");
#endif 

	return ReceiveRecallCommandStatus;
}



 

//**********************************************************************************
//* ReceiveResetResponse - Process the Reset the SPI Slave Controller.  This command 
//* resets the selected controller.  It may be used to implement an emergency stop 
//* function using the PIC's watchdog timer. 
//**********************************************************************************
byte ReceiveResetResponse(void)
{
	byte ResetCommandStatus;
	
#ifndef TEST
    printf("Master: Received Reset Command Response:  ");
  	printf("MessageID = ");
   	dec(CoProcessorMessage.ResetCommand.MessageID); 
   	printf("\r\n\r\n");
#endif

	// Reset the Controller by allowing the Watchdog Timer to timeout.
	
	return ResetCommandStatus;

}

//**********************************************************************************
//* ReceiveGetStatusResponse - Get the latest status from the selected Slave       *
//* controller.  The message includes the controller state, any errors detected and*
//* any diagnostic messages reported.   The current state is returned from the     *
//* function call, while the error and diagnostics codes are saved globally.       *
//**********************************************************************************
byte ReceiveGetStatusResponse(void)
{ 	
	byte GetStatusCommantStatus;
 	
	//CoProcessorMessage.GetStatusCommand.MessageID = MessageIndex++;			     			// Message ID range (0000 - FFFF) Hex
	//CoProcessorMessage.GetStatusCommand.Command = GET_STATUS_COMMAND;		     			// Send Get Status Command
    ControllerState = CoProcessorMessage.GetStatusCommand.ControllerState; 		 			// Get current Controller State
    ControllerErrorCode = CoProcessorMessage.GetStatusCommand.ControllerErrorCode; 			// Get current Controller Error Code
    ControllerDiagnosticCode = CoProcessorMessage.GetStatusCommand.ControllerDiagnosticCode;  // Get current Diagnostic Code
    
#ifndef TEST
    printf("Master: Received Get Status Command Response:  ");
  	printf("MessageID = ");
   	dec(CoProcessorMessage.GetStatusCommand.MessageID);
   	printf("  ControllerState = ");
   	dec(ControllerState);
   	printf("  ControllerErrorCode = ");
   	dec(ControllerErrorCode);
   	printf("  ControllerDiagnosticCode = ");
   	dec(ControllerDiagnosticCode);
   
   	printf("\r\n\r\n");
#endif 
 
	return GetStatusCommantStatus;
	
}


//**********************************************************************************
//* GetResults - Get the results from the latest arithmetic operations             *
//* carried out during the last cycle. Up to 7 separate single precision floating  *
//* point or 3 double precision  values may be returned during a single message    *
//* transfer.                                                                      *
//**********************************************************************************
byte GetResults(float *Results)
{
 	byte GetResultsStatus = TRUE;								// Status return
   	byte AnswerID;									 		    // Local Answer Loop Index 
   
	// Send Get Answers command message to the coprocessor in order to have it 
    // return the floating point answers back to the master.

  	SendGetAnswers();
 				
#ifndef TEST
    printf("master: GetResults: \r\n");
  	printf("  MessageID = ");
   	dec(CoProcessorMessage.GetAnswersCommand.MessageID);
    printf(", Number Of Answers = ");
    dec(CoProcessorMessage.GetAnswersCommand.NumberOfAnswers);  
    printf("\r\n");
#endif

	// Display all the answers in floating point format
	
 	//for (AnswerID = 0; AnswerID < CoProcessorMessage.GetAnswersCommand.NumberOfAnswers; AnswerID++)
 	for (AnswerID = 0; AnswerID < MAX_ANSWERS; AnswerID++)
	{	
		// Copy global Answers to Results array 
		Results[AnswerID] = Answers[AnswerID];

#ifndef TEST
			
  		printf("  Answer ID = ");
    	dec(AnswerID);
    	printf(", Answer = ");
		DoubleToAscii(CoProcessorMessage.GetAnswersCommand.Answers[AnswerID], af1);
   		puts(af1);	      		
		//DoubleToAscii(Results[AnswerID], af1);
   		//hex4(Results[AnswerID]);	              	
    	printf("\r\n");
#endif

	}
		 
	return GetResultsStatus;
}
		

//**********************************************************************************
//* SyncWithSlave - First get the coprocessor's attention by synchronizing with it. 
//* The SYNC byte is sent and an ACK response is expected.  This function is called 
//* before sending or receiving any messages.
//**********************************************************************************
void SyncWithSlave(void)
{
	byte Ch;

	// Issue SYNC byte and wait for the ACK byte from Slave first

 	Ch = 0;
	while (Ch != ACK_BYTE)
	{
		SendByte(SYNC_BYTE);
		Ch = GetByte();
	}

}

//**********************************************************************************
//* SendByte - First get the coprocessor's attention by synchronizing with it. 
//**********************************************************************************

void SendByte(byte TheData)
{
/*
	if(!vUARTIntStatus.UARTIntTxBufferFull)
	{
		UARTIntPutChar(TheData);		
	}
*/
}
//**********************************************************************************
//* GetByte - Read a byte from the serial port
//**********************************************************************************

byte GetByte(void)
{
	byte TheData = 0;
/*
	while(!vUARTIntStatus.UARTIntRxBufferEmpty)
	{
		UARTIntGetChar(&TheData);
	}
*/
	return TheData;
}
 
//******************************************************************************
//* InitializeSPIMaster - Initialize the SPI Master hardware for SPI Clock
//* of Fosc/16, clock Iddle low (SPI Mode 0,0), and sample data after the
//* falling edge of SCK.   
//******************************************************************************
void InitializeSPIMaster(void)
{

  // configure port A for all digital inputs

    //OpenADC(ADC_FOSC_32 &
    //        ADC_RIGHT_JUST & ADC_0ANA_0REF, ADC_CH0 & ADC_CH1 & ADC_CH2 & ADC_CH3 & ADC_CH4 & ADC_CH5 & ADC_CH6 & ADC_CH7 & ADC_INT_OFF);
    //OpenADC(ADC_0ANA_0REF, ADC_INT_OFF);
    OpenADC(ADC_0ANA, ADC_INT_OFF);
 
    // Configure PORTB interrupt with Interrupt on change off and pull-up resistors disabled

    //OpenPORTB(PORTB_CHANGE_INT_OFF & PORTB_PULLUPS_OFF);
    OpenPORTB(PORTB_CHANGE_INT_OFF & PORTB_PULLUPS_ON);
 
    EnablePullups();

	// Configure PORT E for 8 Bits Digital I/O   
 
	PSPCONbits.PSPMODE = 0;

	// Initialize the SPI Master Pin directions
	
	SPI_SCL_DIR = OUTPUT; 					// Set SPI Clock Direction output
 	SPI_SI_DIR = INPUT;						// Set SPI Input Direction input
    SPI_SO_DIR = OUTPUT; 					// Set SPI Output Direction output
	
	SPI_CS_DIR = OUTPUT;					// Set SPI Slave Select direction
	BUSY_FLAG_DIR = INPUT;        			// Set SPI Slave Busy Flag direction           
	SPI_CS = DISABLE; 						// Ensure SPI memory device (disable it first)

	// Chip Select is reset
	//OpenSPI(SPI_FOSC_4, MODE_00, SMPMID);
	//OpenSPI(SPI_FOSC_4, MODE_00, SMPEND);
	
	//OpenSPI(SPI_FOSC_16, MODE_00, SMPMID);
	//OpenSPI(SPI_FOSC_16, MODE_00, SMPEND);

	//OpenSPI(SPI_FOSC_64, MODE_00, SMPMID);
	
	//OpenSPI(SPI_FOSC_64, MODE_00, SMPEND);		//  Select this mode if using Slave Select	(Almost works!)		 

	//OpenSPI(SPI_FOSC_4, MODE_11, SMPMID);

	//OpenSPI(SPI_FOSC_64, MODE_11, SMPEND);
	OpenSPI(SPI_FOSC_64, MODE_11, SMPEND);

	//OpenSPI(SPI_FOSC_64, MODE_11, SMPMID);

	//OpenSPI(SPI_FOSC_4, MODE_01, SMPMID);		// Select this mode if not using Slave Select (SS_BAR) 
	//OpenSPI(SPI_FOSC_4, MODE_01, SMPEND);		// Select this mode if not using Slave Select (SS_BAR) 
	//OpenSPI(SPI_FOSC_16, MODE_01, SMPMID);	// Select this mode if not using Slave Select (SS_BAR) 
	//OpenSPI(SPI_FOSC_16, MODE_01, SMPEND);	// Select this mode if not using Slave Select (SS_BAR) 
	//OpenSPI(SPI_FOSC_64, MODE_01, SMPMID);	// Select this mode if not using Slave Select (SS_BAR) 

	//OpenSPI(SPI_FOSC_64, MODE_01, SMPEND);		// Select this mode if not using Slave Select (SS_BAR) (Almost works!)

	
}

/********************************************************************
*     Function Name:    ReadSPIByte                                 *
*     Return Value:     contents of SSPBUF register                 *
*     Parameters:       void                                        *
*     Description:      Read single byte from SPI bus (Master).     *
********************************************************************/
unsigned char ReadSPIByte( void )
{
  //SSPBUF = 0x00;                  // initiate bus cycle
  while ( !SSPSTATbits.BF );      // wait until cycle complete
  return ( SSPBUF );              // return with byte read 
}

/********************************************************************
*     Function Name:    WriteSPIByte                                *
*     Return Value:     Status byte for WCOL detection.             *
*     Parameters:       Single data byte for SPI bus.               *
*     Description:      This routine writes a single byte to the    * 
*                       SPI bus (Master).                           *
********************************************************************/
unsigned char WriteSPIByte( unsigned char data_out )
{
  SSPBUF = data_out;           // write byte to SSPBUF register
  if ( SSPCON1 & 0x80 )        // test if write collision occurred
   return ( -1 );              // if WCOL bit is set return negative #
  else
  {
    while( !SSPSTATbits.BF );  // wait until bus cycle complete 
  }
  return ( 0 );                // if WCOL bit is not set return non-negative#
}


//******************************************************************************
//* set_wren - Send write enable command (WREN).  
//******************************************************************************
void set_wren(void)
{
	SPI_CS = 0; //assert chip select
	var = putcSPI(SPI_WREN); //send write enable command
	SPI_CS = 1; //negate chip select
}

//******************************************************************************
//* page_write - Send a page write command using the selected address.
//******************************************************************************
void page_write (unsigned char addhigh, 
				 unsigned char addlow, unsigned char *wrptr)
{
	SPI_CS = 0; //assert chip select
	var = putcSPI(SPI_WRITE); //send write command
	var = putcSPI(addhigh); //send high byte of address
	var = putcSPI(addlow); //send low byte of address
	putsSPI(wrptr); //send data byte
	SPI_CS = 1; //negate chip select
}

//******************************************************************************
//* array_read - Read data from the serial EEPROM into an array.  
//******************************************************************************
void array_read (unsigned char addhigh, 
				 unsigned char addlow, unsigned char *rdptr,
				 unsigned char count)
{
	SPI_CS = 0; //assert chip select
	var = putcSPI(SPI_READ); //send read command
	var = putcSPI(addhigh); //send high byte of address
	var = putcSPI(addlow); //send low byte of address
	getsSPI(rdptr, count); //read multiple bytes
	SPI_CS = 1;
}

//******************************************************************************
//* byte_write - Write a byte the serial EEPROM starting at the given address.  
//******************************************************************************
void byte_write (unsigned char addhigh, 
				 unsigned char addlow, unsigned char data)
{
	SPI_CS = 0; //assert chip select
	var = putcSPI(SPI_WRITE); //send write command
	var = putcSPI(addhigh); //send high byte of address
	var = putcSPI(addlow); //send low byte of address
	var = putcSPI(data); //send data byte
	SPI_CS = 1; //negate chip select


}

//******************************************************************************
//* byte_read - Read a byte of data from the serial EEPROM starting at the 
//* given address.  
//******************************************************************************
unsigned char byte_read (unsigned char addhigh,
						 unsigned char addlow)
{
	SPI_CS = 0; //assert chip select
	var = putcSPI(SPI_READ); //send read command
	var = putcSPI(addhigh); //send high byte of address
	var = putcSPI(addlow); //send low byte of address
	var = getcSPI(); //read single byte
	SPI_CS = 1;
	return (var);
}

//******************************************************************************
//* status_read - Read a status byte from the serial EEPROM.
//******************************************************************************
unsigned char status_read (void)
{
	SPI_CS = 0; //assert chip select
	var = putcSPI(SPI_RDSR); //send read status command
	var = getcSPI(); //read data byte
	SPI_CS = 1; //negate chip select
	return (var);
}

//******************************************************************************
//* status_write - Write a status byte to the serial EEPROM.
//******************************************************************************
void status_write (unsigned char data)
{
	SPI_CS = 0;
	var = putcSPI(SPI_WRSR); //write status command
	var = putcSPI(data); //status byte to write
	SPI_CS = 1; //negate chip select
}

//******************************************************************************
//* busy_polling - Wait until the serial EEPROM is not busy anymore.
//******************************************************************************
void busy_polling (void)
{
	do
	{
		SPI_CS = 0; //assert chip select
		var = putcSPI(SPI_RDSR); //send read status command
		var = ReadSPI(); //read data byte
		SPI_CS = 1; //negate chip select
	} while (var & 0x01); //stay in loop until notbusy
}

